
import numpy as np
import os
import platform

from scipy.stats import binom as _binom
from scipy.stats import norm as _norm

from .transform import stft

def median(x, alpha=None, axis=-1, keepdims=False):
    '''
    Computes 95% confidence interval for the median.

    Parameters
    ----------
    x: array_like
        the data array
    alpha: float, optional
        the confidence level of the interval, confidence intervals are only computed
        when this argument is provided
    axis: int, optional
        the axis of the data on which to operate, by default the last axis

    Returns
    -------
    :tuple ``(float, [float, float])``
        This function returns ``(m, [le, ue])`` and the confidence interval is ``[m-le, m+ue]``.
    '''

    # place the axis on which to compute median in first position
    xsw = np.moveaxis(x, axis, 0)

    # sort the array
    xsw = np.sort(xsw, axis=0)
    n = xsw.shape[0]

    if n % 2 == 1:
        # if n is odd, take central element
        m = xsw[n//2,];
    else:
        # if n is even, average the two central elements
        m = 0.5*(xsw[n//2-1,] + xsw[n//2,]);

    if alpha is None:
        if keepdims:
            m = np.moveaxis(m[np.newaxis,], 0, axis)
        return m

    else:
        # bound for using the large n approximation
        clt_bound = max(10 / alpha, 10 / (2 - alpha))

        if n < clt_bound:
            # Get the bounds of the CI from the binomial distribution
            b = _binom(n, 0.5)
            j,k = int(b.ppf(alpha/2)-1), int(b.ppf(1 - alpha/2)-1)

            if b.cdf(k) - b.cdf(j) < 1 - alpha:
                k += 1

            # sanity check
            assert b.cdf(k) - b.cdf(j) >= 1 - alpha

            if j < 0:
                raise ValueError('Warning: Sample size is too small. No confidence interval found.')
            else:
                ci = np.array([xsw[j,]-m, xsw[k,]-m])

        else:
            # we use the Normal approximation for large sets
            norm = _norm()
            eta = norm.ppf(1 - alpha / 2)
            j = int(np.floor(0.5*n - 0.5 * eta * np.sqrt(n))) - 1
            k = int(np.ceil(0.5*n + 0.5 * eta * np.sqrt(n)))
            ci = np.array([xsw[j,]-m,xsw[k,]-m])

        if keepdims:
            m = np.moveaxis(m[np.newaxis,], 0, axis)
            if axis < 0:
                ci = np.moveaxis(ci[:,np.newaxis,], 1, axis)
            else:
                ci = np.moveaxis(ci[:,np.newaxis,], 1, axis+1)


        return m, ci


# Simple mean squared error function
def mse(x1, x2):
    '''
    A short hand to compute the mean-squared error of two signals.

    .. math::

       MSE = \\frac{1}{n}\sum_{i=0}^{n-1} (x_i - y_i)^2


    :arg x1: (ndarray)
    :arg x2: (ndarray)
    :returns: (float) The mean of the squared differences of x1 and x2.
    '''
    return (np.abs(x1-x2)**2).sum()/len(x1)


# Itakura-Saito distance function
def itakura_saito(x1, x2, sigma2_n, stft_L=128, stft_hop=128):

  P1 = np.abs(stft.analysis(x1, stft_L, stft_hop))**2
  P2 = np.abs(stft(x2, stft_L, stft_hop))**2

  VAD1 = P1.mean(axis=1) > 2*stft_L**2*sigma2_n
  VAD2 = P2.mean(axis=1) > 2*stft_L**2*sigma2_n
  VAD = np.logical_or(VAD1, VAD2)

  if P1.shape[0] != P2.shape[0] or P1.shape[1] != P2.shape[1]:
    raise ValueError("Error: Itakura-Saito requires both array to have same length")

  R = P1[VAD,:]/P2[VAD,:]

  IS = (R - np.log(R) - 1.).mean(axis=1)

  return np.median(IS)

def snr(ref, deg):

    return np.sum(ref**2)/np.sum((ref-deg)**2)

# Perceptual Evaluation of Speech Quality for multiple files using multiple threads
def pesq(ref_file, deg_files, Fs=8000, swap=False, wb=False, bin='./bin/pesq'):
    '''
    pesq_vals = pesq(ref_file, deg_files, sample_rate=None, bin='./bin/pesq'):
    Computes the perceptual evaluation of speech quality (PESQ) metric of a degraded
    file with respect to a reference file.  Uses the utility obtained from ITU
    P.862 http://www.itu.int/rec/T-REC-P.862-200511-I!Amd2/en

    :arg ref_file:    The filename of the reference file.
    :arg deg_files:   A list of degraded sound files names.
    :arg sample_rate: Sample rates of the sound files [8kHz or 16kHz, default 8kHz].
    :arg swap:        Swap byte orders (whatever that does is not clear to me) [default: False].
    :arg wb:          Use wideband algorithm [default: False].
    :arg bin:         Location of pesq executable [default: ./bin/pesq].

    :returns: (ndarray size 2xN) ndarray containing Raw MOS and MOS LQO in rows 0 and 1, 
        respectively, and has one column per degraded file name in deg_files.
    '''

    if isinstance(deg_files, str):
        deg_files = [deg_files]

    if platform.system() == 'Windows':
        bin = bin + '.exe'

    if not os.path.isfile(ref_file):
        raise ValueError('Some file did not exist')
    for f in deg_files:
        if not os.path.isfile(f):
            raise ValueError('Some file did not exist')

    if Fs not in (8000, 16000):
        raise ValueError('sample rate must be 8000 or 16000')

    args = [ bin, '+%d' % int(Fs) ]

    if swap is True:
        args.append('+swap')

    if wb is True:
        args.append('+wb')

    args.append(ref_file)

    # array to receive all output values
    pesq_vals = np.zeros((2,len(deg_files)))

    # launch pesq for each degraded file in a different process
    import subprocess
    pipes = [ subprocess.Popen(args+[deg], stdout=subprocess.PIPE) for deg in deg_files ]
    states = np.ones(len(pipes), dtype=np.bool)

    # Recover output as the processes finish
    while states.any():

        for i,p in enumerate(pipes):
            if states[i] == True and p.poll() is not None:
                states[i] = False
                out = p.stdout.readlines()
                last_line = out[-1][:-2]

                if wb is True:
                    if not last_line.startswith('P.862.2 Prediction'):
                        raise ValueError(last_line)
                    pesq_vals[:,i] =  np.array([0, float(last_line.split()[-1])])
                else:
                    if not last_line.startswith('P.862 Prediction'):
                        raise ValueError(last_line)
                    pesq_vals[:,i] = np.array(map(float, last_line.split()[-2:]))

    return pesq_vals



import numpy as np

import pyroomacoustics as pra


def compute_rir(order):
    fromPos = np.zeros((3))
    toPos = np.ones((3, 1))
    roomSize = np.array([3, 3, 3])
    room = pra.ShoeBox(roomSize, fs=1000, absorption=0.95, max_order=order)
    room.add_source(fromPos)
    mics = pra.MicrophoneArray(toPos, room.fs)
    room.add_microphone_array(mics)
    room.compute_rir()


def test_issue_162_max_order_15():
    compute_rir(15)


def test_issue_162_max_order_31():
    compute_rir(31)


def test_issue_162_max_order_32():
    compute_rir(32)


def test_issue_162_max_order_50():
    compute_rir(50)


def test_issue_162_max_order_75():
    compute_rir(75)



# Test of Wall intersection routines
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

eps = 1e-6

cases = {
        '3d_valid' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[-1,-1,-1], [1,1,1]]),
            'intersect' : np.array([1/3,1/3,1/3]),
            'type' : pra.libroom.WALL_ISECT_VALID,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '3d_endpt_1' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[1/3,1/3,1/3], [1,1,1]]),
            'intersect' : np.array([1/3,1/3,1/3]),
            'type' : pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '3d_endpt_2' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[1,1,1], [1/3,1/3,1/3]]),
            'intersect' : np.array([1/3,1/3,1/3]),
            'type' : pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '3d_bndry' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[1,0,1], [-1,0,1]]),
            'intersect' : np.array([0,0,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '3d_vertex' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[-2,1,1], [0,1,1]]),
            'intersect' : np.array([-1,1,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '3d_endpt_bndry_1' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[0,0,1], [-1,0,1]]),
            'intersect' : np.array([0,0,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '3d_endpt_bndry_2' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[1,0,1],[0,0,1]]),
            'intersect' : np.array([0,0,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '3d_vertex_endpt_1' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[-2,1,1], [-1,1,1]]),
            'intersect' : np.array([-1,1,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '3d_vertex_endpt_2' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[-1,1,1], [0,1,1]]),
            'intersect' : np.array([-1,1,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '3d_none_1' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[-10,-5,-3], [-2,-1,-3]]),
            'intersect' : None,
            'type' : pra.libroom.WALL_ISECT_NONE,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '3d_none_2' : {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'seg' : np.array([[-1,-1,2], [1,1,2]]),
            'intersect' : None,
            'type' : pra.libroom.WALL_ISECT_NONE,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '2d_valid' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-1,1], [1,1]]),
            'intersect' : np.array([1/3,1]),
            'type' : pra.libroom.WALL_ISECT_VALID,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '2d_endpt_1' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[1/3,1], [1,1]]),
            'intersect' : np.array([1/3,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '2d_endpt_2' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-1,1], [1/3,1]]),
            'intersect' : np.array([1/3,1]),
            'type' : pra.libroom.WALL_ISECT_VALID_ENDPT,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '2d_bndry_1' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[0,2], [2,2]]),
            'intersect' : np.array([1,2]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '2d_bndry_2' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-2,-1], [0,-1]]),
            'intersect' : np.array([-1,-1]),
            'type' : pra.libroom.WALL_ISECT_VALID_BNDRY,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '2d_bndry_endpt_1' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-1,-1], [0,1]]),
            'intersect' : np.array([-1,-1]),
            'type' : pra.libroom.WALL_ISECT_VALID_ENDPT | pra.libroom.WALL_ISECT_VALID_BNDRY,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '2d_bndry_endpt_2' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-2,-1], [-1,-1]]),
            'intersect' : np.array([-1,-1]),
            'type' : pra.libroom.WALL_ISECT_VALID_ENDPT | pra.libroom.WALL_ISECT_VALID_BNDRY,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        '2d_none_1' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-2,3], [3,3]]),
            'intersect' : None,
            'type' : pra.libroom.WALL_ISECT_NONE,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        '2d_none_2' : {
            'corners' : np.array([  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
                ]),
            'seg' : np.array([[-4,1], [-1,1]]),
            'intersect' : None,
            'type' : pra.libroom.WALL_ISECT_NONE,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },

        }

def run_intersect(lbl):

    case = cases[lbl]
    wall = pra.wall_factory(case['corners'], case['absorption'], case['scattering'])
    p1, p2 = case['seg']

    isect_exp = case['intersect']
    r_exp = case['type']
    isect = np.zeros(wall.dim, dtype=np.float32)
    ret = wall.intersection(p1, p2, isect)

    assert ret == r_exp, 'Wall intersect: expects={} got={}'.format(r_exp, ret)

    if r_exp != pra.libroom.WALL_ISECT_NONE:
        err = np.linalg.norm(isect - case['intersect'])
        assert err < eps, 'Wall intersect: err={} (expects={} got={})'.format(err, isect_exp, isect)


def test_3d_valid():
    run_intersect('3d_valid')

def test_3d_endpt_1():
    run_intersect('3d_endpt_1')

def test_3d_endpt_2():
    run_intersect('3d_endpt_2')

def test_3d_bndry():
    run_intersect('3d_bndry')
    
def test_3d_vertex():
    run_intersect('3d_vertex')

def test_3d_endpt_bndry_1():
    run_intersect('3d_endpt_bndry_1')

def test_3d_endpt_bndry_2():
    run_intersect('3d_endpt_bndry_2')
    
def test_3d_vertex_endpt_1():
    run_intersect('3d_vertex_endpt_1')

def test_3d_vertex_endpt_2():
    run_intersect('3d_vertex_endpt_2')

def test_3d_none_1():
    run_intersect('3d_none_1')

def test_3d_none_2():
    run_intersect('3d_none_2')

def test_2d_valid():
    run_intersect('2d_valid')

def test_2d_endpt_1():
    run_intersect('2d_endpt_1')

def test_2d_endpt_2():
    run_intersect('2d_endpt_2')

def test_2d_bndry_1():
    run_intersect('2d_bndry_1')

def test_2d_bndry_2():
    run_intersect('2d_bndry_2')

def test_2d_bndry_endpt_1():
    run_intersect('2d_bndry_endpt_1')

def test_2d_bndry_endpt_2():
    run_intersect('2d_bndry_endpt_2')

def test_2d_none_1():
    run_intersect('2d_none_1')

def test_2d_none_2():
    run_intersect('2d_none_2')


if __name__ == '__main__':

    test_3d_valid()
    test_3d_endpt_1()
    test_3d_endpt_2()
    test_3d_bndry()
    test_3d_vertex()
    test_3d_endpt_bndry_1()
    test_3d_endpt_bndry_2()
    test_3d_vertex_endpt_1()
    test_3d_vertex_endpt_2()
    test_3d_none_1()
    test_3d_none_2()

    test_2d_valid()
    test_2d_endpt_1()
    test_2d_endpt_2()
    test_2d_bndry_1()
    test_2d_bndry_2()
    test_2d_bndry_endpt_1()
    test_2d_bndry_endpt_2()
    test_2d_none_1()
    test_2d_none_2()




from __future__ import division, print_function

from unittest import TestCase
import numpy as np
from scipy.signal import fftconvolve
import pyroomacoustics as pra

# fix RNG for a deterministic result
np.random.seed(0)

# parameters
length = 15        # the unknown filter length
n_samples = 2000   # the number of samples to run
tol = 1e-2         # tolerance reconstructed filter

# the unknown filter (unit norm)
w = np.random.randn(length)
w /= np.linalg.norm(w)

# create a known driving signal
x = np.random.randn(n_samples)

# convolve with the unknown filter
d_clean = fftconvolve(x, w)[:n_samples]

# a function to the adaptive filter on all the samples
def run_filter(algorithm, x, d):
    for i in range(x.shape[0]):
        algorithm.update(x[i], d[i])

class TestAdaptiveFilter(TestCase):

    def test_rls(self):
        rls = pra.adaptive.RLS(length, lmbd=1., delta=2.0)
        run_filter(rls, x, d_clean)
        error = np.linalg.norm(rls.w - w)
        print('RLS Reconstruction Error', error)
        self.assertTrue(error < tol)

    def test_block_rls(self):
        block_rls = pra.adaptive.BlockRLS(length, lmbd=1., delta=2.0)
        run_filter(block_rls, x, d_clean)
        error = np.linalg.norm(block_rls.w - w)
        print('Block RLS Reconstruction Error', error)
        self.assertTrue(error < tol)

    def test_nlms(self):
        nlms = pra.adaptive.NLMS(length, mu=0.5)
        run_filter(nlms, x, d_clean)
        error = np.linalg.norm(nlms.w - w)
        print('NLMS Reconstruction Error', error)
        self.assertTrue(error < tol)

    def test_block_lms(self):
        block_lms = pra.adaptive.BlockLMS(length, mu=1./length/2.)
        run_filter(block_lms, x, d_clean)
        error = np.linalg.norm(block_lms.w - w)
        print('Block LMS Reconstruction Error', error)
        self.assertTrue(error < tol)




# Author: Eric Bezzam
# Date: July 15, 2016
from __future__ import division, print_function

from .music import *

class CSSM(MUSIC):
    """
    Class to apply the Coherent Signal-Subspace method [CSSM]_ for Direction of
    Arrival (DoA) estimation.

    .. note:: Run locate_sources() to apply the CSSM algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the 
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection 
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate colatitude angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    num_iter: int
        Number of iterations for CSSM. Default: 5

    References
    ----------

    .. [CSSM] H. Wang, M. Kaveh, *Coherent signal-subspace processing for the detection and 
        estimation of angles of arrival of multiple wide-band sources*, IEEE Trans. Acoust., 
        Speech, Signal Process., Vol. 33, Num. 4, pp 823--831, 1985
    """
    def __init__(self, L, fs, nfft, c=343.0, num_src=1, mode='far', r=None,
        azimuth=None, colatitude=None, num_iter=5, **kwargs):

        MUSIC.__init__(self, L=L, fs=fs, nfft=nfft, c=c, num_src=num_src, 
            mode=mode, r=r, azimuth=azimuth, colatitude=colatitude, **kwargs)

        self.iter = num_iter

    def _process(self, X):
        """
        Perform CSSM for given frame in order to estimate steered response 
        spectrum.
        """

        # compute empirical cross correlation matrices
        C_hat = self._compute_correlation_matrices(X)

        # compute initial estimates
        beta = []
        invalid = []

        # Find number of spatial spectrum peaks at each frequency band.
        # If there are less peaks than expected sources, leave the band out
        # Otherwise, store the location of the peaks.
        for k in range(self.num_freq):

            self.grid.set_values(1 / self._compute_spatial_spectrum(C_hat[k,:,:],
                                 self.freq_bins[k]))
            idx = self.grid.find_peaks(k=self.num_src)

            if len(idx) < self.num_src:    # remove frequency
                invalid.append(k)
            else:
                beta.append(idx)

        # Here we remove the bands that had too few peaks
        self.freq_bins = np.delete(self.freq_bins, invalid)
        self.num_freq = self.num_freq - len(invalid)

        # compute reference frequency (take bin with max amplitude)
        f0 = np.argmax(np.sum(np.sum(abs(X[:,self.freq_bins,:]), axis=0),
            axis=1))
        f0 = self.freq_bins[f0]

        # iterate to find DOA, maximum number of iterations is 20
        i = 0

        # while(i < self.iter or (len(self.src_idx) < self.num_src and i < 20)):
        while(i < self.iter):

            # coherent sum
            R = self._coherent_sum(C_hat, f0, beta)

            # subspace decomposition
            Es, En, ws, wn = self._subspace_decomposition(R)

            # compute spatial spectrum
            cross = np.dot(En,np.conjugate(En).T)

            # cross = np.identity(self.M) - np.dot(Es, np.conjugate(Es).T)
            self.grid.set_values(self._compute_spatial_spectrum(cross,f0))
            idx = self.grid.find_peaks(k=self.num_src)
            beta = np.tile(idx, (self.num_freq, 1))

            i += 1

    def _coherent_sum(self, C_hat, f0, beta):

        R = np.zeros((self.M,self.M))

        # coherently sum frequencies
        for j in range(len(self.freq_bins)):
            k = self.freq_bins[j]

            Aj = self.mode_vec[k,:,beta[j]].T
            A0 = self.mode_vec[f0,:,beta[j]].T

            B = np.concatenate((np.zeros([self.M-len(beta[j]), len(beta[j])]), 
                np.identity(self.M-len(beta[j]))), axis=1).T

            Tj = np.dot(np.c_[A0, B], np.linalg.inv(np.c_[Aj, B]))

            R = R + np.dot(np.dot(Tj,C_hat[j,:,:]),np.conjugate(Tj).T)

        return R



'''
Routines to perform grid search on the sphere
'''
from __future__ import division, print_function, absolute_import

import numpy as np
import scipy.spatial as sp #import ConvexHull, SphericalVoronoi

from abc import ABCMeta, abstractmethod

from .detect_peaks import detect_peaks
from .utils import great_circ_dist

class Grid:
    '''
    This is an abstract class with attributes and methods for grids

    Parameters
    ----------
    n_points: int
        the number of points on the grid
    '''

    __metaclass__ = ABCMeta

    def __init__(self, n_points):

        self.n_points = n_points
        self.values = None
        self.cartesian = np.zeros((3,n_points))
        self.spherical = np.zeros((2,n_points))
        self.x = self.cartesian[0,:]
        self.y = self.cartesian[1,:]
        self.z = self.cartesian[2,:]
        self.azimuth = self.spherical[0,:]
        self.colatitude = self.spherical[1,:]
        self.dim = 0
        self.values = None

    @abstractmethod
    def apply(self, func, spherical=False):
        return NotImplemented

    @abstractmethod
    def find_peaks(self, k=1):
        return NotImplemented

    def set_values(self, vals):

        vals = np.array(vals)

        if vals.ndim == 0:
            self.values = np.ones(self.n_points) * vals

        else:
            if vals.shape != (self.n_points,):
                raise ValueError('Values should be a scalar or a 1D ndarray of the grid size.')

            self.values = vals


class GridCircle(Grid):
    '''
    Creates a grid on the circle.

    Parameters
    ----------
    n_points: int, optional
        The number of uniformly spaced points in the grid.
    azimuth: ndarray, optional
        An array of azimuth (in radians) to use for grid locations. Overrides n_points.
    '''

    def __init__(self, n_points=360, azimuth=None):
        if azimuth is not None:

            if azimuth.ndim != 1:
                raise ValueError("Azimuth must be a 1D ndarray.")

            azimuth = np.sort(azimuth)

            n_points = azimuth.shape[0]

        else:
            azimuth = np.linspace(0, 2*np.pi, n_points, endpoint=False)

        Grid.__init__(self, n_points)

        self.dim = 2

        # spherical coordinates
        self.azimuth[:] = azimuth
        self.colatitude[:] = np.pi / 2  # Fix colatitude to ecuador

        # cartesian coordinates
        self.x[:] = np.cos(azimuth)
        self.y[:] = np.sin(azimuth)

    def apply(self, func, spherical=False):

        if spherical:
            self.values = func(self.azimuth)
        else:
            self.values = func(self.x, self.y)

    def find_peaks(self, k=1):

        # make circular
        val_ext = np.append(self.values ,self.values[:10])

        # run peak finding
        indexes = detect_peaks(val_ext, show=False) % self.n_points
        candidates = np.unique(indexes)  # get rid of duplicates, if any

        # Select k largest
        peaks = self.values[candidates]
        max_idx = np.argsort(peaks)[-k:]

        # return the indices of peaks found
        return candidates[max_idx]

    def plot(self, mark_peaks=0):

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings
            warnings.warn('Matplotlib is required for plotting')
            return

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='polar')

        pts = np.append(self.azimuth, self.azimuth[0])
        vals = np.append(self.values, self.values[0])

        ax.plot(pts, vals, '-')

        if mark_peaks > 0:

            idx = self.find_peaks(k=mark_peaks)

            ax.plot(pts[idx], vals[idx], 'ro')


class GridSphere(Grid):
    '''
    This function computes nearly equidistant points on the sphere
    using the fibonacci method

    Parameters
    ----------
    n_points: int
        The number of points to sample
    spherical_points: ndarray, optional
        A 2 x n_points array of spherical coordinates with azimuth in
        the top row and colatitude in the second row. Overrides n_points.

    References
    ----------
    http://lgdv.cs.fau.de/uploads/publications/spherical_fibonacci_mapping.pdf
    http://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere
    '''

    def __init__(self, n_points=1000, spherical_points=None):
        if spherical_points is not None:

            if spherical_points.ndim != 2 or spherical_points.shape[0] != 2:
                raise ValueError('spherical_points must be a 2D array with two rows.')

            n_points = spherical_points.shape[1]

        # Parent constructor
        Grid.__init__(self, n_points)

        self.dim = 3

        if spherical_points is not None:
            # If a list of points was provided, use it

            self.spherical[:,:] = spherical_points

            # transform to cartesian coordinates
            self.x[:] = np.cos(self.azimuth) * np.sin(self.colatitude)
            self.y[:] = np.sin(self.azimuth) * np.sin(self.colatitude)
            self.z[:] = np.cos(self.colatitude)

        else:
            # If no list was provided, samples points on the sphere
            # as uniformly as possible

            # Fibonnaci sampling
            offset = 2. / n_points
            increment = np.pi * (3. - np.sqrt(5.))

            self.z[:] = (np.arange(n_points) * offset - 1) + offset / 2
            rho = np.sqrt(1. - self.z**2)

            phi = np.arange(n_points) * increment

            self.x[:] = np.cos(phi) * rho
            self.y[:] = np.sin(phi) * rho

            # Create convenient arrays
            # to access both in cartesian and spherical coordinates
            self.azimuth[:] = np.arctan2(self.y, self.x)
            self.colatitude[:] = np.arctan2(np.sqrt(self.x**2 + self.y**2), self.z)

        # To perform the peak detection in 2D on a non-squared grid it is
        # necessary to know the neighboring points of each grid point.  The
        # Convex Hull of points on the sphere is equivalent to the Delauney
        # triangulation of the point set, which is what we are looking for.

        # Now we also want to compute the convex hull
        self.hull = sp.ConvexHull(self.cartesian.T)

        # and create an adjacency list
        adjacency = [ set() for pt in range(self.n_points) ]

        # Simplices contains all the triangles that form
        # the faces of the convex hull. We use this to find which
        # points are connected on the hull.
        for tri in self.hull.simplices:
            adjacency[tri[0]].add(tri[1])
            adjacency[tri[0]].add(tri[2])
            adjacency[tri[1]].add(tri[0])
            adjacency[tri[1]].add(tri[2])
            adjacency[tri[2]].add(tri[0])
            adjacency[tri[2]].add(tri[1])

        # convert to list of lists
        self.neighbors = [ list(x) for x in adjacency ]


    def apply(self, func, spherical=False):
        '''
        Apply a function to every grid point
        '''

        if spherical:
            self.values = func(self.azimuth, self.colatitude)
        else:
            self.values = func(self.x, self.y, self.z)


    def min_max_distance(self):
        ''' Compute some statistics on the distribution of the points '''

        min_dist = np.inf
        max_dist = 0

        dist = []

        for u in range(self.n_points):

            phi1, theta1 = self.spherical[:,u]

            for v in self.neighbors[u]:

                phi2, theta2 = self.spherical[:,v]

                d = great_circ_dist(1, theta1, phi1, theta2, phi2)

                dist.append(d)

                if d < min_dist:
                    min_dist = d
                
                if d > max_dist:
                    max_dist = d

        dist = np.array(dist)
        mean = dist.mean()
        median = np.median(dist)

        return min_dist, max_dist, mean, median


    def find_peaks(self, k=1):
        '''
        Find the largest peaks on the grid
        '''
        
        candidates = []

        # We start by looking at points whose neighbors all have lower values
        # than themselves
        for v in range(self.n_points):

            is_local_max = True

            for u in self.neighbors[v]:
                if self.values[u] > self.values[v]:
                    is_local_max = False
                    break

            if is_local_max:
                candidates.append(v)

        # Now sort and return k largest
        I = np.argsort(self.values[candidates])

        return [candidates[x] for x in I[-k:]]


    def regrid(self):
        ''' Regrid the non-uniform data on a regular mesh '''

        if self.values is None:
            warnings.warn('Cannont regrid: data missing.')
            return

        # First we need to interpolate the non-uniformly sampled data
        # onto a uniform grid
        from scipy.interpolate import griddata
        
        x = int(np.sqrt(self.n_points / 2))

        G = np.mgrid[-np.pi:np.pi:2j*x, 0:np.pi:1j*x]
        spherical_grid = np.squeeze(G.reshape((2,1,-1)))

        gridded = griddata(self.spherical.T, self.values, spherical_grid.T, method='nearest', fill_value=0.)
        dirty_img = gridded.reshape((2*x, x))

        return G[0], G[1], gridded.reshape((2*x, x))


    def plot(self, 
            colatitude_ref=None, azimuth_ref=None,
            colatitude_recon=None, azimuth_recon=None,
            plotly=True, projection=True, points_only=False):

        if points_only:
            dirty_img = None
            dirty_grid_x = None
            dirty_grid_y = None
        else:
            dirty_grid_x, dirty_grid_y, dirty_img = self.regrid()

        ## Then, we just need to call Hanjie's routines
        from .plotters import sph_plot_diracs, sph_plot_diracs_plotly

        if projection:
            sph_plot_diracs(
                    colatitude_ref, azimuth_ref, colatitude_recon, azimuth_recon, 
                    dirty_img=dirty_img,
                    colatitude_grid=dirty_grid_y, azimuth_grid=dirty_grid_x,
                    )

        if plotly:
            sph_plot_diracs_plotly(
                    colatitude_ref, azimuth_ref, colatitude_recon, azimuth_recon, 
                    dirty_img=dirty_img,
                    azimuth_grid=dirty_grid_x, colatitude_grid=dirty_grid_y
                    )

    
    def plot_old(self, plot_points=False, mark_peaks=0):
        ''' Plot the points on the sphere with their values '''

        from scipy import rand

        try:
            import matplotlib.colors as colors
            #from mpl_toolkits.mplot3d import Axes3D
            import mpl_toolkits.mplot3d as a3
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings
            warnings.warn('Matplotlib is required for plotting')
            return

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        if plot_points:
            ax.scatter(self.x, self.y, self.z, c='b', marker='o')

        if mark_peaks > 0:

            id = self.find_peaks(k=mark_peaks)
            s = 1.05
            ax.scatter(s*self.x[id], s*self.y[id], s*self.z[id], c='k', marker='o')

        voronoi = sp.SphericalVoronoi(self.cartesian.T)
        voronoi.sort_vertices_of_regions()

        if self.values is not None:

            col_max = self.values.max()
            col_min = self.values.min()

            if col_min != col_max:
                col_map = (self.values - col_min) / (col_max - col_min)
            else:
                col_map = self.values / col_max

        else:

            col_map = np.zeros(self.n_points)

        cmap = plt.get_cmap('coolwarm')

        # plot the walls
        for v_ind, col in zip(voronoi.regions, col_map):
            triangle = a3.art3d.Poly3DCollection([voronoi.vertices[v_ind]], alpha=1., linewidth=0.0)
            triangle.set_color(cmap(col))
            triangle.set_edgecolor('k')
            ax.add_collection3d(triangle)

        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        ax.set_xlim([-1,1])
        ax.set_ylim([-1,1])
        ax.set_zlim([-1,1])

